@@ -20,6 +20,7 @@ struct cmd_info {
   char *args;
   char *help;
   int is_real_cmd;
+  int is_option;
 };
 
 static struct cmd_info command_table[] = {
@@ -30,7 +31,8 @@ struct cmd_node {
   struct cmd_node *sibling, *son, **plastson;
   struct cmd_info *cmd, *help;
   int len;
-  signed char prio;
+  u8 final;
+  s8 prio;
   char token[1];
 };
 
@@ -51,20 +53,26 @@ cmd_build_tree(void)
       struct cmd_node *old, *new;
       char *c = cmd->command;
 
-      old = &cmd_root;
+      new = &cmd_root;
       while (*c)
 	{
+	  /* Walk through the keywords in the command from start of the command
+	   * to its end */
 	  char *d = c;
 	  while (*c && !isspace_(*c))
 	    c++;
+	  old = new;
 	  for(new=old->son; new; new=new->sibling)
 	    if (new->len == c-d && !memcmp(new->token, d, c-d))
 	      break;
 	  if (!new)
 	    {
+	      /* If new keyword is not stored yet */
 	      int size = sizeof(struct cmd_node) + c-d;
 	      new = malloc(size);
 	      bzero(new, size);
+	      /* Making this keyword child of the previons
+	       * keyword in the command */
 	      *old->plastson = new;
 	      old->plastson = &new->sibling;
 	      new->plastson = &new->son;
@@ -72,14 +80,19 @@ cmd_build_tree(void)
 	      memcpy(new->token, d, c-d);
 	      new->prio = (new->len == 3 && (!memcmp(new->token, "roa", 3) || !memcmp(new->token, "rip", 3))) ? 0 : 1; /* Hack */
 	    }
-	  old = new;
 	  while (isspace_(*c))
 	    c++;
 	}
+
+      /* In new there is now stored the last keyword in command.
+       * Only this command is going to contain cmd and help */
       if (cmd->is_real_cmd)
-	old->cmd = cmd;
+	new->cmd = cmd;
       else
-	old->help = cmd;
+	new->help = cmd;
+
+      if (cmd->is_option)
+	old->final = 1;
     }
 }
 
@@ -93,12 +106,23 @@ cmd_do_display_help(struct cmd_info *c)
 }
 
 static void
-cmd_display_help(struct cmd_info *c1, struct cmd_info *c2)
+cmd_display_help(struct cmd_node *m)
 {
-  if (c1)
-    cmd_do_display_help(c1);
-  else if (c2)
-    cmd_do_display_help(c2);
+  /* Go through children of the base keyword. Each keyword has its own node. */
+  for (struct cmd_node *mm = m->son; mm; mm = mm->sibling)
+  {
+    /* Print hints for all of the children of given keyword. */
+    if (mm->help)
+      cmd_do_display_help(mm->help);
+    else if (mm->cmd)
+      cmd_do_display_help(mm->cmd);
+    else
+    {
+      /* This child does not contain hint, let's try its children instead
+       * (this child might be first part of multi keyword command) */
+      cmd_display_help(mm);
+    }
+  }
 }
 
 static struct cmd_node *
@@ -135,7 +159,7 @@ cmd_list_ambiguous(struct cmd_node *root, char *cmd, int len)
 
   for(m=root->son; m; m=m->sibling)
     if (m->len > len && !memcmp(m->token, cmd, len))
-      cmd_display_help(m->help, m->cmd);
+      cmd_display_help(m);
 }
 
 void
@@ -147,7 +171,7 @@ cmd_help(char *cmd, int len)
   int ambig;
 
   n = &cmd_root;
-  while (cmd < end)
+  while (cmd < end && !n->final)
     {
       if (isspace_(*cmd))
 	{
@@ -167,9 +191,15 @@ cmd_help(char *cmd, int len)
 	break;
       n = m;
     }
-  cmd_display_help(n->cmd, NULL);
-  for (m=n->son; m; m=m->sibling)
-    cmd_display_help(m->help, m->cmd);
+
+  if (n && n->cmd)
+    cmd_do_display_help(n->cmd);
+
+  /* Currently no help for options */
+  if (n->final)
+    return;
+
+  cmd_display_help(n);
 }
 
 static int
@@ -229,7 +259,7 @@ cmd_complete(char *cmd, int len, char *buf, int again)
 
   /* Find the context */
   n = &cmd_root;
-  while (cmd < fin && n->son)
+  while (cmd < fin && n->son && !n->final)
     {
       if (isspace_(*cmd))
 	{
@@ -290,7 +320,7 @@ cmd_expand(char *cmd)
 
   args = c = cmd;
   n = &cmd_root;
-  while (*c)
+  while (*c && !n->final)
     {
       if (isspace_(*c))
 	{
