@@ -29,7 +29,6 @@
  */
 
 #include "proto/bmp/bmp.h"
-#include "proto/bmp/buffer.h"
 #include "proto/bmp/map.h"
 
 #include <sys/socket.h>
@@ -94,54 +93,26 @@ enum bmp_message_type {
   BMP_ROUTE_MIRROR_MSG = 6 // Route Mirroring Message
 };
 
-// Total size of Common Header
-#define BMP_COMMON_HDR_SIZE 6
-// Defines size of padding when IPv4 address is going to be put into field
-// which can accept also IPv6 address
-#define BMP_PADDING_IP4_ADDR_SIZE 12
-
-/* BMP Per-Peer Header [RFC 7854 - Section 4.2] */
-// Total size of Per-Peer Header
-#define BMP_PER_PEER_HDR_SIZE 42
-
 enum bmp_peer_type {
   BMP_PEER_TYPE_GLOBAL_INSTANCE = 0,
   BMP_PEER_TYPE_RD_INSTANCE = 1,
   BMP_PEER_TYPE_LOCAL_INSTANCE = 2
 };
 
-#define BMP_PEER_HDR_FLAG_V_SHIFT 7
-enum bmp_peer_flag_v_t {
-  // The Peer address is an IPv4 address
-  BMP_PEER_HDR_FLAG_V_IP4 = (0 << BMP_PEER_HDR_FLAG_V_SHIFT),
-  // The Peer address is an IPv6 address
-  BMP_PEER_HDR_FLAG_V_IP6 = (1 << BMP_PEER_HDR_FLAG_V_SHIFT)
+#define BMP_PEER_FLAG_V_IPV6		(1 << 7)
+#define BMP_PEER_FLAG_L_POLICY		(1 << 6)
+#define BMP_PEER_FLAG_A_NO_AS4		(1 << 5)
+
+struct bmp_peer_hdr_info {
+  ip_addr address;
+  u32 as;
+  u32 id;
+  bool global;
+  bool policy;
+  bool no_as4;
+  btime timestamp;
 };
 
-#define BMP_PEER_HDR_FLAG_L_SHIFT 6
-enum bmp_peer_flag_l {
-  BMP_PEER_HDR_FLAG_L_PRE_POLICY_ADJ_RIB_IN = (0 << BMP_PEER_HDR_FLAG_L_SHIFT),
-  BMP_PEER_HDR_FLAG_L_POST_POLICY_ADJ_RIB_IN = (1 << BMP_PEER_HDR_FLAG_L_SHIFT)
-};
-
-#define BMP_PEER_HDR_FLAG_A_SHIFT 5
-enum bmp_peer_flag_a {
-  // The 4-byte AS_PATH format
-  BMP_PEER_HDR_FLAG_A_AS_PATH_4B = (0 << BMP_PEER_HDR_FLAG_A_SHIFT),
-  // The legacy 2-byte AS_PATH format
-  BMP_PEER_HDR_FLAG_A_AS_PATH_2B = (1 << BMP_PEER_HDR_FLAG_A_SHIFT)
-};
-
-#define BMP_PEER_HDR_FLAGS_INIT(flags) \
-  (flags) = 0
-#define BMP_PEER_HDR_FLAGS_SET(flags, bit_mask) \
-  (flags) |= (bit_mask)
-
-/* BMP Information TLV header [RFC 7854 - Section 4.4] */
-// Total size of Type and Length fields of Information TLV Header without
-// variable part
-#define BMP_INFO_TLV_FIX_SIZE 4
-
 enum bmp_info_tlv_type {
   BMP_INFO_TLV_TYPE_STRING = 0,    // String
   BMP_INFO_TLV_TYPE_SYS_DESCR = 1, // SysDescr
@@ -166,15 +137,22 @@ enum bmp_peer_down_notif_reason {
   BMP_PEER_DOWN_REASON_PEER_DE_CONFIGURED = 5
 };
 
+struct bmp_peer_down_info {
+  u8 reason;
+  u8 fsm_code;
+  u8 err_code;
+  u8 err_subcode;
+  const byte *data;
+  int length;
+};
+
 /* BMP Termination Message [RFC 7854 - Section 4.5] */
-#define BMP_TERM_INFO_TYPE_SIZE 2
 enum bmp_term_info_type {
   BMP_TERM_INFO_STRING = 0, // The Information field contains string
   BMP_TERM_INFO_REASON = 1, // The Information field contains 2-byte reason code
 };
 
 // 2-byte code in the Information field
-#define BMP_TERM_REASON_CODE_SIZE 2
 enum bmp_term_reason {
   BMP_TERM_REASON_ADM = 0,  // Session administratively closed
   BMP_TERM_REASON_UNK = 1,  // Unspecified reason
@@ -183,12 +161,6 @@ enum bmp_term_reason {
   BMP_TERM_REASON_PERM = 4, // Session permanently administratively closed
 };
 
-// Size of Information Length field in Termination Message header
-#define BMP_TERM_INFO_LEN_FIELD_SIZE 2
-
-// Default chunk size request when memory allocation
-#define DEFAULT_MEM_BLOCK_SIZE 4096
-
 // Initial delay for connection to the BMP collector
 #define CONNECT_INIT_TIME (200 MS)
 
@@ -198,21 +170,42 @@ enum bmp_term_reason {
 #define IP4_MAX_TTL 255
 
 
-#define IF_COND_TRUE_PRINT_ERR_MSG_AND_RETURN_OPT_VAL(expr, msg, rv...)     \
-  do {                                                                      \
-    if ((expr))                                                             \
-    {                                                                       \
-      log(L_WARN "[BMP] " msg);                                             \
-      return rv;                                                            \
-    }                                                                       \
-  } while (0)
+#define bmp_buffer_need(b, sz)  ASSERT_DIE((b)->pos + (sz) <= (b)->end)
 
+// Idea for following macros has been taken from |proto/mrt/mrt.c|
+#define BMP_DEFINE_PUT_FUNC(S, T)                               \
+  static inline void                                            \
+  bmp_put_##S(buffer *b, const T x)                             \
+  {                                                             \
+    bmp_buffer_need(b, sizeof(T));			   \
+    put_##S(b->pos, x);                                    \
+    b->pos += sizeof(T);                                   \
+  }
 
-#define IF_PTR_IS_NULL_PRINT_ERR_MSG_AND_RETURN_OPT_VAL(p, msg, rv...)	\
-  do {									\
-    IF_COND_TRUE_PRINT_ERR_MSG_AND_RETURN_OPT_VAL(!(p), msg, rv);	\
-  } while (0)
+BMP_DEFINE_PUT_FUNC(u8, u8)
+BMP_DEFINE_PUT_FUNC(u16, u16)
+BMP_DEFINE_PUT_FUNC(u32, u32)
+BMP_DEFINE_PUT_FUNC(u64, u64)
+BMP_DEFINE_PUT_FUNC(ip4, ip4_addr)
+BMP_DEFINE_PUT_FUNC(ip6, ip6_addr)
 
+static inline void
+bmp_put_data(buffer *b, const void *src, const size_t len)
+{
+  ASSERT_DIE(b->pos + len <= b->end);
+  memcpy(b->pos, src, len);
+  b->pos += len;
+}
+
+static inline buffer
+bmp_default_buffer(struct bmp_proto *p)
+{
+  return (buffer) {
+    .start = p->msgbuf,
+    .pos = p->msgbuf,
+    .end = p->msgbuf + sizeof p->msgbuf,
+  };
+}
 
 static void bmp_connected(struct birdsock *sk);
 static void bmp_sock_err(sock *sk, int err);
@@ -225,35 +218,69 @@ bmp_send_peer_up_notif_msg(struct bmp_proto *p, const struct bgp_proto *bgp,
 
 static void bmp_route_monitor_end_of_rib(struct bmp_proto *p, struct bmp_stream *bs);
 
-// Stores necessary any data in list
-struct bmp_data_node {
-  node n;
-  byte *data;
-  size_t data_size;
+// Stores TX data
+struct bmp_tx_buffer {
+  struct bmp_tx_buffer *next;
+  byte *pos;
+  byte data[];
+};
 
-  u32 remote_as;
-  u32 remote_id;
-  ip_addr remote_ip;
-  btime timestamp;
-  bool global_peer;
-  bool policy;
+#define bmp_tx_remains(b) (((byte *) (b) + page_size) - (b)->pos)
+
+/* A dummy resource to accurately show memory pages allocated for pending TX */
+struct bmp_tx_resource {
+  resource r;
+  struct bmp_proto *p;
 };
 
 static void
-bmp_common_hdr_serialize(buffer *stream, const enum bmp_message_type type, const u32 data_size)
+bmp_tx_resource_free(resource *r UNUSED) {}
+
+static void
+bmp_tx_resource_dump(struct dump_request *dreq UNUSED, resource *r UNUSED) {}
+
+static struct resmem
+bmp_tx_resource_memsize(resource *r)
 {
+  struct bmp_proto *p = SKIP_BACK(struct bmp_tx_resource, r, r)->p;
+
+  return (struct resmem) {
+    .effective = p->tx_pending_count * page_size,
+    .overhead = sizeof(struct bmp_tx_resource),
+  };
+}
+
+static struct resclass bmp_tx_resource_class = {
+  .name = "BMP TX buffers",
+  .size = sizeof(struct bmp_tx_resource),
+  .free = bmp_tx_resource_free,
+  .dump = bmp_tx_resource_dump,
+  .memsize = bmp_tx_resource_memsize,
+};
+
+static void
+bmp_put_common_hdr(buffer *stream, enum bmp_message_type type, u32 length)
+{
+  bmp_buffer_need(stream, BMP_COMMON_HDR_SIZE);
   bmp_put_u8(stream, BMP_VERSION_3);
-  bmp_put_u32(stream, BMP_COMMON_HDR_SIZE + data_size);
+  bmp_put_u32(stream, length);
   bmp_put_u8(stream, type);
 }
 
 static void
-bmp_info_tlv_hdr_serialize(buffer *stream, const enum bmp_info_tlv_type type,
-  const char *str)
+bmp_fix_common_hdr(buffer *stream)
+{
+  uint length = stream->pos - stream->start;
+  put_u32(stream->start + 1, length);
+}
+
+static void
+bmp_put_info_tlv(buffer *stream, enum bmp_info_tlv_type type, const char *str)
 {
   size_t str_len = strlen(str);
   str_len = MIN(str_len, MIB_II_STR_LEN);
 
+  bmp_buffer_need(stream, 4 + str_len);
   bmp_put_u16(stream, type);
   bmp_put_u16(stream, str_len);
   bmp_put_data(stream, str, str_len);
@@ -263,33 +290,72 @@ bmp_info_tlv_hdr_serialize(buffer *stream, const enum bmp_info_tlv_type type,
 static void
 bmp_init_msg_serialize(buffer *stream, const char *sys_descr, const char *sys_name)
 {
-  const size_t sys_descr_len = strlen(sys_descr);
-  const size_t sys_name_len = strlen(sys_name);
-  // We include MIB-II sysDescr and sysName in BMP INIT MSG so that's why
-  // allocated 2x BMP_INFO_TLV_FIX_SIZE memory pool size
-  const size_t data_size = (2 * BMP_INFO_TLV_FIX_SIZE) + sys_descr_len + sys_name_len;
-  bmp_buffer_need(stream, BMP_COMMON_HDR_SIZE + data_size);
-  bmp_common_hdr_serialize(stream, BMP_INIT_MSG, data_size);
-  bmp_info_tlv_hdr_serialize(stream, BMP_INFO_TLV_TYPE_SYS_DESCR, sys_descr);
-  bmp_info_tlv_hdr_serialize(stream, BMP_INFO_TLV_TYPE_SYS_NAME, sys_name);
+  bmp_put_common_hdr(stream, BMP_INIT_MSG, 0);
+  bmp_put_info_tlv(stream, BMP_INFO_TLV_TYPE_SYS_DESCR, sys_descr);
+  bmp_put_info_tlv(stream, BMP_INFO_TLV_TYPE_SYS_NAME, sys_name);
+  bmp_fix_common_hdr(stream);
 }
 
 static void
-bmp_schedule_tx_packet(struct bmp_proto *p, const byte *payload, const size_t size)
+bmp_schedule_tx_packet(struct bmp_proto *p, buffer *msg)
 {
+  const byte *payload = msg->start;
+  size_t size = msg->pos - msg->start;
+
   ASSERT(p->started);
 
-  struct bmp_data_node *tx_data = mb_alloc(p->tx_mem_pool, sizeof (struct bmp_data_node));
-  tx_data->data = mb_alloc(p->tx_mem_pool, size);
-  memcpy(tx_data->data, payload, size);
-  tx_data->data_size = size;
-  add_tail(&p->tx_queue, &tx_data->n);
-
-  if (sk_tx_buffer_empty(p->sk)
-      && !ev_active(p->tx_ev))
+  while (size)
   {
-    ev_schedule(p->tx_ev);
+    if (!p->tx_last || !bmp_tx_remains(p->tx_last))
+    {
+      if (p->tx_pending_count >= p->tx_pending_limit)
+	return ev_schedule(p->tx_overflow_event);
+
+      p->tx_pending_count++;
+
+      struct bmp_tx_buffer *btb = alloc_page();
+      btb->pos = btb->data;
+      btb->next = NULL;
+
+      if (p->tx_last)
+      {
+	ASSERT_DIE(!p->tx_last->next);
+	p->tx_last->next = btb;
+      }
+      else
+	ASSERT_DIE(p->tx_pending_count == 1);
+
+      p->tx_last = btb;
+
+      if (!p->tx_pending)
+	p->tx_pending = btb;
+    }
+
+    size_t cpylen = bmp_tx_remains(p->tx_last);
+    if (size < cpylen)
+      cpylen = size;
+
+    memcpy(p->tx_last->pos, payload, cpylen);
+    p->tx_last->pos += cpylen;
+
+    payload += cpylen;
+    size -= cpylen;
   }
+
+  if (!p->sk->tbuf && !ev_active(p->tx_ev))
+    ev_schedule(p->tx_ev);
+}
+
+static void
+bmp_tx_buffer_free(struct bmp_proto *p, struct bmp_tx_buffer *btb)
+{
+  if (btb == p->tx_last)
+  {
+    p->tx_last = NULL;
+    ASSERT_DIE(!p->tx_pending_count);
+  }
+
+  free_page(btb);
 }
 
 static void
@@ -300,40 +366,33 @@ bmp_fire_tx(void *p_)
   if (!p->started)
     return;
 
-  IF_COND_TRUE_PRINT_ERR_MSG_AND_RETURN_OPT_VAL(
-    EMPTY_LIST(p->tx_queue),
-    "Called BMP TX event handler when there is not any data to send"
-  );
-
-  size_t cnt = 0; // Counts max packets which we want to send per TX slot
-  struct bmp_data_node *tx_data;
-  struct bmp_data_node *tx_data_next;
-  WALK_LIST_DELSAFE(tx_data, tx_data_next, p->tx_queue)
+  int cnt = 0;
+  for (struct bmp_tx_buffer *btb; btb = p->tx_pending; )
   {
-    if (tx_data->data_size > p->sk->tbsize)
-    {
-      sk_set_tbsize(p->sk, tx_data->data_size);
-    }
+    ASSERT_DIE(!p->sk->tbuf);
 
-    size_t data_size = tx_data->data_size;
-    memcpy(p->sk->tbuf, tx_data->data, data_size);
-    mb_free(tx_data->data);
-    rem_node((node *) tx_data);
-    mb_free(tx_data);
+    p->sk->tbuf = btb->data;
+    u64 sz = btb->pos - btb->data;
 
-    if (sk_send(p->sk, data_size) <= 0)
+    p->tx_sent += sz;
+    p->tx_sent_total += sz;
+
+    if (p->tx_pending == p->tx_last)
+      p->tx_last = NULL;
+
+    p->tx_pending = btb->next;
+    p->tx_pending_count--;
+
+    if (sk_send(p->sk, sz) <= 0)
       return;
 
-    // BMP packets should be treat with lowest priority when scheduling sending
-    // packets to target. That's why we want to send max. 32 packets per event
-    // call
-    if (++cnt > 32)
+    p->sk->tbuf = NULL;
+    bmp_tx_buffer_free(p, btb);
+
+    if (cnt++ > 1024)
     {
       if (!ev_active(p->tx_ev))
-      {
-        ev_schedule(p->tx_ev);
-      }
-
+	ev_schedule(p->tx_ev);
       return;
     }
   }
@@ -342,6 +401,13 @@ bmp_fire_tx(void *p_)
 static void
 bmp_tx(struct birdsock *sk)
 {
+  struct bmp_proto *p = sk->data;
+
+  struct bmp_tx_buffer *btb = SKIP_BACK(struct bmp_tx_buffer, data, sk->tbuf);
+  bmp_tx_buffer_free(p, btb);
+
+  sk->tbuf = NULL;
+
   bmp_fire_tx(sk->data);
 }
 
@@ -374,7 +440,7 @@ bmp_put_bgp_hdr(buffer *stream, const u8 msg_type, const u16 msg_length)
 }
 
 /**
- * bmp_per_peer_hdr_serialize - serializes Per-Peer Header
+ * bmp_put_per_peer_hdr - serializes Per-Peer Header
  *
  * @is_post_policy: indicate the message reflects the post-policy Adj-RIB-In
  * @peer_addr: the remote IP address associated with the TCP session
@@ -384,75 +450,42 @@ bmp_put_bgp_hdr(buffer *stream, const u8 msg_type, const u16 msg_length)
  * @ts_usec: the time in microseconds when the encapsulated routes were received
  */
 static void
-bmp_per_peer_hdr_serialize(buffer *stream, const bool is_global_instance_peer,
-  const bool is_post_policy, const bool is_as_path_4bytes,
-  const ip_addr peer_addr, const u32 peer_as, const u32 peer_bgp_id,
-  const u32 ts_sec, const u32 ts_usec)
+bmp_put_per_peer_hdr(buffer *stream, const struct bmp_peer_hdr_info *peer)
 {
-  // TODO: ATM we don't support BMP_PEER_TYPE_RD_INSTANCE
-  const enum bmp_peer_type peer_type = is_global_instance_peer
-                                      ? BMP_PEER_TYPE_GLOBAL_INSTANCE
-                                      : BMP_PEER_TYPE_LOCAL_INSTANCE;
-  const u8 peer_flag_v = ipa_is_ip4(peer_addr)
-                           ? BMP_PEER_HDR_FLAG_V_IP4
-                           : BMP_PEER_HDR_FLAG_V_IP6;
-  const u8 peer_flag_l = is_post_policy
-                           ? BMP_PEER_HDR_FLAG_L_POST_POLICY_ADJ_RIB_IN
-                           : BMP_PEER_HDR_FLAG_L_PRE_POLICY_ADJ_RIB_IN;
-  const u8 peer_flag_a = is_as_path_4bytes
-                           ? BMP_PEER_HDR_FLAG_A_AS_PATH_4B
-                           : BMP_PEER_HDR_FLAG_A_AS_PATH_2B;
-  u8 peer_flags;
-  BMP_PEER_HDR_FLAGS_INIT(peer_flags);
-  BMP_PEER_HDR_FLAGS_SET(peer_flags, peer_flag_v);
-  BMP_PEER_HDR_FLAGS_SET(peer_flags, peer_flag_l);
-  BMP_PEER_HDR_FLAGS_SET(peer_flags, peer_flag_a);
+  // TODO: ATM we do not support BMP_PEER_TYPE_RD_INSTANCE
+  u8 peer_type = peer->global ?
+    BMP_PEER_TYPE_GLOBAL_INSTANCE :
+    BMP_PEER_TYPE_LOCAL_INSTANCE;
 
+  u8 peer_flags =
+    (ipa_is_ip6(peer->address) ? BMP_PEER_FLAG_V_IPV6 : 0) |
+    (peer->policy ? BMP_PEER_FLAG_L_POLICY : 0) |
+    (peer->no_as4 ? BMP_PEER_FLAG_A_NO_AS4 : 0);
+
+  u32 ts_sec = peer->timestamp TO_S;
+  u32 ts_usec = peer->timestamp - (ts_sec S);
+
+  bmp_buffer_need(stream, BMP_PER_PEER_HDR_SIZE);
   bmp_put_u8(stream, peer_type);
   bmp_put_u8(stream, peer_flags);
-  // TODO: Provide appropriate peer Route Distinguisher if applicable
-  bmp_put_u64(stream, 0x00); // 0x00 - Not supported peer distinguisher
-  bmp_put_ipa(stream, peer_addr);
-  bmp_put_u32(stream, peer_as);
-  bmp_put_u32(stream, peer_bgp_id);
+  bmp_put_u64(stream, 0); // Not supported peer distinguisher
+  bmp_put_ipa(stream, peer->address);
+  bmp_put_u32(stream, peer->as);
+  bmp_put_u32(stream, peer->id);
   bmp_put_u32(stream, ts_sec);
   bmp_put_u32(stream, ts_usec);
 }
 
-/* [4.6] Route Monitoring */
 static void
-bmp_route_monitor_msg_serialize(buffer *stream, const bool is_peer_global,
-  const bool table_in_post_policy, const u32 peer_as, const u32 peer_bgp_id,
-  const bool as4_support, const ip_addr remote_addr, const byte *update_msg,
-  const size_t update_msg_size, btime timestamp)
+bmp_peer_up_notif_msg_serialize(buffer *stream, const struct bmp_peer_hdr_info *peer,
+				const ip_addr local_addr, const u16 local_port, const u16 remote_port,
+				const byte *sent_msg, const size_t sent_msg_length,
+				const byte *recv_msg, const size_t recv_msg_length)
 {
-  const size_t data_size = BMP_PER_PEER_HDR_SIZE + update_msg_size;
-  u32 ts_sec = timestamp TO_S;
-  u32 ts_usec = timestamp - (ts_sec S);
+  bmp_put_common_hdr(stream, BMP_PEER_UP_NOTIF, 0);
+  bmp_put_per_peer_hdr(stream, peer);
 
-  bmp_buffer_need(stream, BMP_COMMON_HDR_SIZE + data_size);
-  bmp_common_hdr_serialize(stream, BMP_ROUTE_MONITOR, data_size);
-  bmp_per_peer_hdr_serialize(stream, is_peer_global, table_in_post_policy,
-    as4_support, remote_addr, peer_as, peer_bgp_id, ts_sec, ts_usec);
-  bmp_put_data(stream, update_msg, update_msg_size);
-}
-
-static void
-bmp_peer_up_notif_msg_serialize(buffer *stream, const bool is_peer_global,
-  const u32 peer_as, const u32 peer_bgp_id, const bool as4_support,
-  const ip_addr local_addr, const ip_addr remote_addr, const u16 local_port,
-  const u16 remote_port, const byte *sent_msg, const size_t sent_msg_length,
-  const byte *recv_msg, const size_t recv_msg_length)
-{
-  const size_t data_size =
-    BMP_PER_PEER_HDR_SIZE + BMP_PEER_UP_NOTIF_MSG_FIX_SIZE +
-    BGP_HEADER_LENGTH + sent_msg_length + BGP_HEADER_LENGTH + recv_msg_length;
-
-  bmp_buffer_need(stream, BMP_COMMON_HDR_SIZE + data_size);
-  bmp_common_hdr_serialize(stream, BMP_PEER_UP_NOTIF, data_size);
-  bmp_per_peer_hdr_serialize(stream, is_peer_global,
-    false /* TODO: Hardcoded pre-policy Adj-RIB-In */, as4_support, remote_addr,
-    peer_as, peer_bgp_id, 0, 0); // 0, 0 - No timestamp provided
+  bmp_buffer_need(stream, BMP_PEER_UP_NOTIF_MSG_FIX_SIZE);
   bmp_put_ipa(stream, local_addr);
   bmp_put_u16(stream, local_port);
   bmp_put_u16(stream, remote_port);
@@ -460,20 +493,37 @@ bmp_peer_up_notif_msg_serialize(buffer *stream, const bool is_peer_global,
   bmp_put_data(stream, sent_msg, sent_msg_length);
   bmp_put_bgp_hdr(stream, PKT_OPEN, BGP_HEADER_LENGTH + recv_msg_length);
   bmp_put_data(stream, recv_msg, recv_msg_length);
+
+  bmp_fix_common_hdr(stream);
 }
 
 static void
-bmp_peer_down_notif_msg_serialize(buffer *stream, const bool is_peer_global,
-  const u32 peer_as, const u32 peer_bgp_id, const bool as4_support,
-  const ip_addr remote_addr, const byte *data, const size_t data_size)
+bmp_peer_down_notif_msg_serialize(buffer *stream, const struct bmp_peer_hdr_info *peer,
+				  const struct bmp_peer_down_info *info)
 {
-  const size_t payload_size = BMP_PER_PEER_HDR_SIZE + data_size;
-  bmp_buffer_need(stream, BMP_COMMON_HDR_SIZE + payload_size);
-  bmp_common_hdr_serialize(stream, BMP_PEER_DOWN_NOTIF, payload_size);
-  bmp_per_peer_hdr_serialize(stream, is_peer_global,
-    false /* TODO: Hardcoded pre-policy adj RIB IN */,  as4_support, remote_addr,
-    peer_as, peer_bgp_id, 0, 0); // 0, 0 - No timestamp provided
-  bmp_put_data(stream, data, data_size);
+  bmp_put_common_hdr(stream, BMP_PEER_DOWN_NOTIF, 0);
+  bmp_put_per_peer_hdr(stream, peer);
+
+  bmp_put_u8(stream, info->reason);
+
+  switch (info->reason)
+  {
+  case BMP_PEER_DOWN_REASON_LOCAL_BGP_NOTIFICATION:
+  case BMP_PEER_DOWN_REASON_REMOTE_BGP_NOTIFICATION:;
+    uint bgp_msg_length = BGP_HEADER_LENGTH + 2 + info->length;
+    bmp_buffer_need(stream, bgp_msg_length);
+    bmp_put_bgp_hdr(stream, PKT_NOTIFICATION, bgp_msg_length);
+    bmp_put_u8(stream, info->err_code);
+    bmp_put_u8(stream, info->err_subcode);
+    bmp_put_data(stream, info->data, info->length);
+    break;
+
+  case BMP_PEER_DOWN_REASON_LOCAL_NO_NOTIFICATION:
+    bmp_put_u16(stream, info->fsm_code);
+    break;
+  }
+
+  bmp_fix_common_hdr(stream);
 }
 
 
@@ -765,125 +815,114 @@ bmp_send_peer_up_notif_msg(struct bmp_proto *p, const struct bgp_proto *bgp,
   ASSERT(p->started);
 
   const struct birdsock *sk = bmp_get_birdsock_ext(bgp);
-  IF_PTR_IS_NULL_PRINT_ERR_MSG_AND_RETURN_OPT_VAL(
-    sk,
-    "[BMP] No BGP socket"
-  );
+  if (!sk)
+  {
+    log(L_WARN "%s: No BGP socket", p->p.name);
+    return;
+  }
 
-  const bool is_global_instance_peer = bmp_is_peer_global_instance(bgp);
-  buffer payload = bmp_buffer_alloc(p->buffer_mpool, DEFAULT_MEM_BLOCK_SIZE);
-  bmp_peer_up_notif_msg_serialize(&payload, is_global_instance_peer,
-    bgp->remote_as, bgp->remote_id, 1,
-    sk->saddr, sk->daddr, sk->sport, sk->dport, tx_data, tx_data_size,
-    rx_data, rx_data_size);
-  bmp_schedule_tx_packet(p, bmp_buffer_data(&payload), bmp_buffer_pos(&payload));
-  bmp_buffer_free(&payload);
-}
+  struct bmp_peer_hdr_info peer = {
+    .address = bgp->remote_ip,
+    .as = bgp->remote_as,
+    .id = bgp->remote_id,
+    .global = bmp_is_peer_global_instance(bgp),
+    .policy = false,		// Hardcoded pre-policy Adj-RIB-In
+    .timestamp = 0,		// No timestamp provided
+  };
 
-static void
-bmp_route_monitor_put_update(struct bmp_proto *p, struct bmp_stream *bs, const byte *data, size_t length, btime timestamp)
-{
-  struct bmp_data_node *upd_msg = mb_alloc(p->update_msg_mem_pool,
-                               sizeof (struct bmp_data_node));
-  upd_msg->data = mb_alloc(p->update_msg_mem_pool, length);
-  memcpy(upd_msg->data, data, length);
-  upd_msg->data_size = length;
-  add_tail(&p->update_msg_queue, &upd_msg->n);
-
-  /* Save some metadata */
-  struct bgp_proto *bgp = bs->bgp;
-  upd_msg->remote_as = bgp->remote_as;
-  upd_msg->remote_id = bgp->remote_id;
-  upd_msg->remote_ip = bgp->remote_ip;
-  upd_msg->timestamp = timestamp;
-  upd_msg->global_peer = bmp_is_peer_global_instance(bgp);
-  upd_msg->policy = bmp_stream_policy(bs);
-
-  /* Kick the commit */
-  if (!ev_active(p->update_ev))
-    ev_schedule(p->update_ev);
+  buffer payload = bmp_default_buffer(p);
+  bmp_peer_up_notif_msg_serialize(&payload, &peer, sk->saddr, sk->sport, sk->dport,
+				  tx_data, tx_data_size, rx_data, rx_data_size);
+  bmp_schedule_tx_packet(p, &payload);
 }
 
 static void
 bmp_route_monitor_notify(struct bmp_proto *p, struct bmp_stream *bs,
 			 const net_addr *n, const struct rte *new, const struct rte_src *src)
 {
-  byte buf[BGP_MAX_EXT_MSG_LENGTH];
-  byte *end = bgp_bmp_encode_rte(bs->sender, buf, n, new, src);
+  struct bgp_proto *bgp = bs->bgp;
+  struct bgp_channel *c = bs->sender;
 
   btime delta_t = new ? current_time() - new->lastmod : 0;
-  btime timestamp = current_real_time() - delta_t;
 
-  if (end)
-    bmp_route_monitor_put_update(p, bs, buf, end - buf, timestamp);
-  else
-    log(L_WARN "%s: Cannot encode update for %N", p->p.name, n);
-}
+  struct bmp_peer_hdr_info peer = {
+    .address = bgp->remote_ip,
+    .as = bgp->remote_as,
+    .id = bgp->remote_id,
+    .global = bmp_is_peer_global_instance(bgp),
+    .policy = bmp_stream_policy(bs),
+    .timestamp = current_real_time() - delta_t,
+  };
 
-static void
-bmp_route_monitor_commit(void *p_)
-{
-  struct bmp_proto *p = p_;
+  buffer msg = bmp_default_buffer(p);
+  bmp_put_common_hdr(&msg, BMP_ROUTE_MONITOR, 0);
+  bmp_put_per_peer_hdr(&msg, &peer);
 
-  if (!p->started)
-    return;
-
-  buffer payload
-    = bmp_buffer_alloc(p->buffer_mpool, DEFAULT_MEM_BLOCK_SIZE);
-
-  struct bmp_data_node *data, *data_next;
-  WALK_LIST_DELSAFE(data, data_next, p->update_msg_queue)
+  bmp_buffer_need(&msg, BGP_MAX_EXT_MSG_LENGTH);
+  byte *pos = bgp_bmp_encode_rte(c, msg.pos + BGP_HEADER_LENGTH, msg.end, n, new, src);
+  if (!pos)
   {
-    bmp_route_monitor_msg_serialize(&payload,
-      data->global_peer, data->policy,
-      data->remote_as, data->remote_id, true,
-      data->remote_ip, data->data, data->data_size,
-      data->timestamp);
-
-    bmp_schedule_tx_packet(p, bmp_buffer_data(&payload), bmp_buffer_pos(&payload));
-
-    bmp_buffer_flush(&payload);
-
-    mb_free(data->data);
-    rem_node(&data->n);
-    mb_free(data);
+    log(L_WARN "%s: Cannot encode update for %N", p->p.name, n);
+    return;
   }
+  bmp_put_bgp_hdr(&msg, PKT_UPDATE, pos - msg.pos);
+  msg.pos = pos;
 
-  bmp_buffer_free(&payload);
+  bmp_fix_common_hdr(&msg);
+  bmp_schedule_tx_packet(p, &msg);
 }
 
 static void
 bmp_route_monitor_end_of_rib(struct bmp_proto *p, struct bmp_stream *bs)
 {
-  TRACE(D_PACKETS, "Sending END-OF-RIB for %s.%s", bs->bgp->p.name, bs->sender->c.name);
+  struct bgp_proto *bgp = bs->bgp;
+  struct bgp_channel *c = bs->sender;
 
-  byte rx_end_payload[DEFAULT_MEM_BLOCK_SIZE];
-  byte *pos = bgp_create_end_mark_(bs->sender, rx_end_payload + BGP_HEADER_LENGTH);
-  memset(rx_end_payload + BGP_MSG_HDR_MARKER_POS, 0xff,
-	 BGP_MSG_HDR_MARKER_SIZE); // BGP UPDATE MSG marker
-  put_u16(rx_end_payload + BGP_MSG_HDR_LENGTH_POS, pos - rx_end_payload);
-  put_u8(rx_end_payload + BGP_MSG_HDR_TYPE_POS, PKT_UPDATE);
+  struct bmp_peer_hdr_info peer = {
+    .address = bgp->remote_ip,
+    .as = bgp->remote_as,
+    .id = bgp->remote_id,
+    .global = bmp_is_peer_global_instance(bgp),
+    .policy = bmp_stream_policy(bs),
+    .timestamp = current_real_time(),
+  };
 
-  bmp_route_monitor_put_update(p, bs, rx_end_payload, pos - rx_end_payload, current_real_time());
+  TRACE(D_PACKETS, "Sending END-OF-RIB for %s.%s", bgp->p.name, c->c.name);
+
+  buffer msg = bmp_default_buffer(p);
+  bmp_put_common_hdr(&msg, BMP_ROUTE_MONITOR, 0);
+  bmp_put_per_peer_hdr(&msg, &peer);
+
+  bmp_buffer_need(&msg, BGP_MAX_EXT_MSG_LENGTH);
+  byte *pos = bgp_create_end_mark_(c, msg.pos + BGP_HEADER_LENGTH);
+  bmp_put_bgp_hdr(&msg, PKT_UPDATE, pos - msg.pos);
+  msg.pos = pos;
+
+  bmp_fix_common_hdr(&msg);
+  bmp_schedule_tx_packet(p, &msg);
 }
 
 static void
 bmp_send_peer_down_notif_msg(struct bmp_proto *p, const struct bgp_proto *bgp,
-  const byte *data, const size_t data_size)
+			     const struct bmp_peer_down_info *info)
 {
   ASSERT(p->started);
 
   const struct bgp_caps *remote_caps = bmp_get_bgp_remote_caps_ext(bgp);
-  bool is_global_instance_peer = bmp_is_peer_global_instance(bgp);
-  buffer payload
-    = bmp_buffer_alloc(p->buffer_mpool, DEFAULT_MEM_BLOCK_SIZE);
-  bmp_peer_down_notif_msg_serialize(&payload, is_global_instance_peer,
-    bgp->remote_as, bgp->remote_id,
-    remote_caps ? remote_caps->as4_support : bgp->as4_session,
-    bgp->remote_ip, data, data_size);
-  bmp_schedule_tx_packet(p, bmp_buffer_data(&payload), bmp_buffer_pos(&payload));
 
-  bmp_buffer_free(&payload);
+  struct bmp_peer_hdr_info peer = {
+    .address = bgp->remote_ip,
+    .as = bgp->remote_as,
+    .id = bgp->remote_id,
+    .global = bmp_is_peer_global_instance(bgp),
+    .policy = false,		// Hardcoded pre-policy Adj-RIB-In
+    .no_as4 = remote_caps ? !remote_caps->as4_support : !bgp->as4_session,
+    .timestamp = 0,		// No timestamp provided
+  };
+
+  buffer payload = bmp_default_buffer(p);
+  bmp_peer_down_notif_msg_serialize(&payload, &peer, info);
+  bmp_schedule_tx_packet(p, &payload);
 }
 
 static void
@@ -899,48 +938,32 @@ bmp_peer_down_(struct bmp_proto *p, const struct bgp_proto *bgp,
 
   TRACE(D_STATES, "Peer down for %s", bgp->p.name);
 
-  uint bmp_code = 0;
-  uint fsm_code = 0;
+  struct bmp_peer_down_info info = {
+    .err_code = err_code,
+    .err_subcode = err_subcode,
+    .data = data,
+    .length = length,
+  };
 
   switch (err_class)
   {
   case BE_BGP_RX:
-    bmp_code = BMP_PEER_DOWN_REASON_REMOTE_BGP_NOTIFICATION;
+    info.reason = BMP_PEER_DOWN_REASON_REMOTE_BGP_NOTIFICATION;
     break;
 
   case BE_BGP_TX:
   case BE_AUTO_DOWN:
   case BE_MAN_DOWN:
-    bmp_code = BMP_PEER_DOWN_REASON_LOCAL_BGP_NOTIFICATION;
+    info.reason = BMP_PEER_DOWN_REASON_LOCAL_BGP_NOTIFICATION;
     break;
 
   default:
-    bmp_code = BMP_PEER_DOWN_REASON_REMOTE_NO_NOTIFICATION;
-    length = 0;
+    info.reason = BMP_PEER_DOWN_REASON_REMOTE_NO_NOTIFICATION;
+    info.length = 0;
     break;
   }
 
-  buffer payload = bmp_buffer_alloc(p->buffer_mpool, 1 + BGP_HEADER_LENGTH + 2 + length);
-  bmp_put_u8(&payload, bmp_code);
-
-  switch (bmp_code)
-  {
-  case BMP_PEER_DOWN_REASON_LOCAL_BGP_NOTIFICATION:
-  case BMP_PEER_DOWN_REASON_REMOTE_BGP_NOTIFICATION:
-    bmp_put_bgp_hdr(&payload, BGP_HEADER_LENGTH + 2 + length, PKT_NOTIFICATION);
-    bmp_put_u8(&payload, err_code);
-    bmp_put_u8(&payload, err_subcode);
-    bmp_put_data(&payload, data, length);
-    break;
-
-  case BMP_PEER_DOWN_REASON_LOCAL_NO_NOTIFICATION:
-    bmp_put_u16(&payload, fsm_code);
-    break;
-  }
-
-  bmp_send_peer_down_notif_msg(p, bgp, bmp_buffer_data(&payload), bmp_buffer_pos(&payload));
-
-  bmp_buffer_free(&payload);
+  bmp_send_peer_down_notif_msg(p, bgp, &info);
 
   bmp_remove_peer(p, bp);
 }
@@ -955,25 +978,24 @@ bmp_peer_down(const struct bgp_proto *bgp,
 }
 
 static void
-bmp_send_termination_msg(struct bmp_proto *p,
-  const enum bmp_term_reason reason)
+bmp_send_termination_msg(struct bmp_proto *p, enum bmp_term_reason reason)
 {
-  const size_t term_msg_hdr_size = BMP_TERM_INFO_TYPE_SIZE
-                                     + BMP_TERM_INFO_LEN_FIELD_SIZE
-                                     + BMP_TERM_REASON_CODE_SIZE;
-  const size_t term_msg_size = BMP_COMMON_HDR_SIZE + term_msg_hdr_size;
-  buffer stream = bmp_buffer_alloc(p->buffer_mpool, term_msg_size);
-  bmp_common_hdr_serialize(&stream, BMP_TERM_MSG, term_msg_hdr_size);
-  bmp_put_u16(&stream, BMP_TERM_INFO_REASON);
-  bmp_put_u16(&stream, BMP_TERM_REASON_CODE_SIZE); // 2-byte code indication the reason
-  bmp_put_u16(&stream, reason);
-  memcpy(p->sk->tbuf, bmp_buffer_data(&stream), bmp_buffer_pos(&stream));
-  IF_COND_TRUE_PRINT_ERR_MSG_AND_RETURN_OPT_VAL(
-    sk_send(p->sk, bmp_buffer_pos(&stream)) < 0,
-    "Failed to send BMP termination message"
-    );
+  buffer stream = bmp_default_buffer(p);
 
-  bmp_buffer_free(&stream);
+  bmp_put_common_hdr(&stream, BMP_TERM_MSG, 0);
+  bmp_buffer_need(&stream, 6);
+  bmp_put_u16(&stream, BMP_TERM_INFO_REASON);
+  bmp_put_u16(&stream, 2); // Length of the reason
+  bmp_put_u16(&stream, reason);
+  bmp_fix_common_hdr(&stream);
+
+  if (p->sk->tbuf)
+    bmp_tx_buffer_free(p, SKIP_BACK(struct bmp_tx_buffer, data, p->sk->tbuf));
+
+  p->sk->tbuf = stream.start;
+  if (sk_send(p->sk, stream.pos - stream.start) < 0)
+    log(L_WARN "%s: Cannot send BMP termination message", p->p.name);
+  p->sk->tbuf = NULL;
 }
 
 int
@@ -1062,10 +1084,9 @@ bmp_startup(struct bmp_proto *p)
   proto_notify_state(&p->p, PS_UP);
 
   /* Send initiation message */
-  buffer payload = bmp_buffer_alloc(p->buffer_mpool, DEFAULT_MEM_BLOCK_SIZE);
+  buffer payload = bmp_default_buffer(p);
   bmp_init_msg_serialize(&payload, p->sys_descr, p->sys_name);
-  bmp_schedule_tx_packet(p, bmp_buffer_data(&payload), bmp_buffer_pos(&payload));
-  bmp_buffer_free(&payload);
+  bmp_schedule_tx_packet(p, &payload);
 
   /* Send Peer Up messages */
   struct proto *peer;
@@ -1086,6 +1107,7 @@ bmp_down(struct bmp_proto *p)
 {
   ASSERT(p->started);
   p->started = false;
+  p->tx_sent = 0;
 
   TRACE(D_EVENTS, "BMP session closed");
 
@@ -1119,7 +1141,6 @@ bmp_connect(struct bmp_proto *p)
   sk->dport = p->station_port;
   sk->ttl = IP4_MAX_TTL;
   sk->tos = IP_PREC_INTERNET_CONTROL;
-  sk->tbsize = BGP_TX_BUFFER_EXT_SIZE;
   sk->tx_hook = bmp_connected;
   sk->err_hook = bmp_sock_err;
 
@@ -1173,6 +1194,26 @@ bmp_sock_err(sock *sk, int err)
   proto_notify_state(&p->p, PS_START);
 }
 
+static void
+bmp_tx_overflow(void *_p)
+{
+  struct bmp_proto *p = _p;
+  if (p->tx_pending_count < p->tx_pending_limit)
+    return;
+
+  p->sock_err = 0;
+
+  log(L_ERR "%s: Connection stalled", p->p.name);
+
+  if (p->started)
+    bmp_down(p);
+
+  bmp_close_socket(p);
+  tm_start(p->connect_retry_timer, CONNECT_RETRY_TIME);
+
+  proto_notify_state(&p->p, PS_START);
+}
+
 /* BMP connect timeout event - switch from Idle/Connect state to Connect state */
 static void
 bmp_connection_retry(timer *t)
@@ -1189,6 +1230,24 @@ bmp_connection_retry(timer *t)
 static void
 bmp_close_socket(struct bmp_proto *p)
 {
+  if (p->sk && p->sk->tbuf)
+    bmp_tx_buffer_free(p, SKIP_BACK(struct bmp_tx_buffer, data, p->sk->tbuf));
+
+  struct bmp_tx_buffer *btb = p->tx_pending;
+  while (btb)
+  {
+    p->tx_pending_count--;
+
+    struct bmp_tx_buffer *next = btb->next;
+    bmp_tx_buffer_free(p, btb);
+    btb = next;
+  }
+
+  p->tx_pending = NULL;
+
+  ASSERT_DIE(!p->tx_last);
+  ASSERT_DIE(!p->tx_pending_count);
+
   rfree(p->sk);
   p->sk = NULL;
 }
@@ -1230,6 +1289,7 @@ bmp_init(struct proto_config *CF)
   strcpy(p->sys_name, cf->sys_name);
   p->monitoring_rib.in_pre_policy = cf->monitoring_rib_in_pre_policy;
   p->monitoring_rib.in_post_policy = cf->monitoring_rib_in_post_policy;
+  p->tx_pending_limit = cf->tx_pending_limit;
 
   return P;
 }
@@ -1243,21 +1303,20 @@ bmp_start(struct proto *P)
 {
   struct bmp_proto *p = (void *) P;
 
-  p->buffer_mpool = rp_new(P->pool, "BMP Buffer");
-  p->map_mem_pool = rp_new(P->pool, "BMP Map");
-  p->tx_mem_pool = rp_new(P->pool, "BMP Tx");
-  p->update_msg_mem_pool = rp_new(P->pool, "BMP Update");
   p->tx_ev = ev_new_init(p->p.pool, bmp_fire_tx, p);
-  p->update_ev = ev_new_init(p->p.pool, bmp_route_monitor_commit, p);
+  p->tx_pending = NULL;
+  p->tx_pending_count = 0;
+  p->tx_overflow_event = ev_new_init(p->p.pool, bmp_tx_overflow, p);
   p->connect_retry_timer = tm_new_init(p->p.pool, bmp_connection_retry, p, 0, 0);
   p->sk = NULL;
 
+  resource *r = ralloc(P->pool, &bmp_tx_resource_class);
+  SKIP_BACK(struct bmp_tx_resource, r, r)->p = p;
+
   HASH_INIT(p->peer_map, P->pool, 4);
   HASH_INIT(p->stream_map, P->pool, 4);
   HASH_INIT(p->table_map, P->pool, 4);
 
-  init_list(&p->tx_queue);
-  init_list(&p->update_msg_queue);
   p->started = false;
   p->sock_err = 0;
   add_tail(&bmp_proto_list, &p->bmp_node);
@@ -1276,6 +1335,7 @@ bmp_shutdown(struct proto *P)
   {
     bmp_send_termination_msg(p, BMP_TERM_REASON_ADM);
     bmp_down(p);
+    bmp_close_socket(p);
   }
 
   p->sock_err = 0;
@@ -1306,6 +1366,9 @@ bmp_reconfigure(struct proto *P, struct proto_config *CF)
   /* We must update our copy of configuration ptr */
   p->cf = new;
 
+  /* Reconfigure tx buffer size limits */
+  p->tx_pending_limit = new->tx_pending_limit;
+
   return 1;
 }
 
@@ -1342,6 +1405,13 @@ bmp_show_proto_info(struct proto *P)
 
     if (p->sock_err)
       cli_msg(-1006, "  %-19s %M", "Last error:", p->sock_err);
+
+    cli_msg(-1006, "  %-19s % 9sB (limit %sB)", "Pending TX:",
+	fmt_order(p->tx_pending_count * (u64) page_size, 1, 10000),
+	fmt_order(p->tx_pending_limit * (u64) page_size, 1, 10000));
+
+    cli_msg(-1006, "  %-19s % 9sB", "Session TX:", fmt_order(p->tx_sent, 1, 10000));
+    cli_msg(-1006, "  %-19s % 9sB", "Total TX:", fmt_order(p->tx_sent_total, 1, 10000));
   }
 }
 
