@@ -594,7 +594,10 @@ mrt_table_dump_free(struct mrt_table_dump_state *s)
 static int
 mrt_table_dump_step(struct mrt_table_dump_state *s)
 {
-  struct bgp_write_state bws = { .as4_session = 1 };
+  struct bgp_write_state bws = {
+    .as4_session = 1,
+    .ignore_non_bgp_attrs = 1,
+  };
 
   s->max = 2048;
   s->bws = &bws;
