@@ -65,31 +65,19 @@ pipe_rt_notify(struct proto *P, struct channel *src_ch, net *n, rte *new, rte *o
 
   if (new)
     {
+      src = new->src;
+
       a = alloca(rta_size(new->attrs));
       memcpy(a, new->attrs, rta_size(new->attrs));
 
-      a->aflags = 0;
+      a->cached = 0;
       a->hostentry = NULL;
-      e = rte_get_temp(a);
-      e->pflags = 0;
-
-      /* Copy protocol specific embedded attributes. */
-      memcpy(&(e->u), &(new->u), sizeof(e->u));
-      e->pref = new->pref;
-      e->pflags = new->pflags;
-
-#ifdef CONFIG_BGP
-      /* Hack to cleanup cached value */
-      if (e->attrs->src->proto->proto == &proto_bgp)
-	e->u.bgp.stale = -1;
-#endif
-
-      src = a->src;
+      e = rte_get_temp(a, src);
     }
   else
     {
       e = NULL;
-      src = old->attrs->src;
+      src = old->src;
     }
 
   src_ch->table->pipe_busy = 1;
@@ -98,11 +86,11 @@ pipe_rt_notify(struct proto *P, struct channel *src_ch, net *n, rte *new, rte *o
 }
 
 static int
-pipe_preexport(struct proto *P, rte **ee, struct linpool *p UNUSED)
+pipe_preexport(struct channel *C, rte *e)
 {
-  struct proto *pp = (*ee)->sender->proto;
+  struct proto *pp = e->sender->proto;
 
-  if (pp == P)
+  if (pp == C->proto)
     return -1;	/* Avoid local loops automatically */
 
   return 0;
@@ -304,3 +292,9 @@ struct protocol proto_pipe = {
   .get_status = 	pipe_get_status,
   .show_proto_info = 	pipe_show_proto_info
 };
+
+void
+pipe_build(void)
+{
+  proto_build(&proto_pipe);
+}
